package ru.dbotthepony.kstarbound.lua.bindings

import com.google.gson.JsonObject
import com.google.gson.reflect.TypeToken
import org.classdump.luna.ByteString
import org.classdump.luna.LuaRuntimeException
import org.classdump.luna.Table
import ru.dbotthepony.kommons.gson.set
import ru.dbotthepony.kstarbound.Registries
import ru.dbotthepony.kstarbound.Starbound
import ru.dbotthepony.kstarbound.defs.DamageData
import ru.dbotthepony.kstarbound.defs.EphemeralStatusEffect
import ru.dbotthepony.kstarbound.defs.actor.PersistentStatusEffect
import ru.dbotthepony.kstarbound.fromJsonFast
import ru.dbotthepony.kstarbound.lua.LuaEnvironment
import ru.dbotthepony.kstarbound.lua.from
import ru.dbotthepony.kstarbound.lua.get
import ru.dbotthepony.kstarbound.lua.iterator
import ru.dbotthepony.kstarbound.lua.luaFunction
import ru.dbotthepony.kstarbound.lua.set
import ru.dbotthepony.kstarbound.lua.tableOf
import ru.dbotthepony.kstarbound.lua.toByteString
import ru.dbotthepony.kstarbound.lua.toJsonFromLua
import ru.dbotthepony.kstarbound.util.sbIntern
import ru.dbotthepony.kstarbound.world.entities.StatusController
import kotlin.math.absoluteValue

private object PersistentStatusEffectToken : TypeToken<PersistentStatusEffect>()
private object CPersistentStatusEffectToken : TypeToken<ArrayList<PersistentStatusEffect>>()

fun provideStatusControllerBindings(self: StatusController, lua: LuaEnvironment) {
	val callbacks = lua.newTable()
	lua.globals["status"] = callbacks

	callbacks["statusProperty"] = luaFunction { key: ByteString, ifMissing: Any? ->
		val get = self.getProperty(key.decode())

		if (get == null) {
			returnBuffer.setTo(ifMissing)
		} else {
			returnBuffer.setTo(from(get))
		}
	}

	callbacks["setStatusProperty"] = luaFunction { key: ByteString, value: Any? ->
		self.setProperty(key.decode().sbIntern(), toJsonFromLua(value))
	}

	callbacks["stat"] = luaFunction { name: ByteString ->
		returnBuffer.setTo(self.liveStats[name.decode()]?.effectiveModifiedValue ?: 0.0)
	}

	callbacks["statPositive"] = luaFunction { name: ByteString ->
		returnBuffer.setTo(self.statPositive(name.decode()))
	}

	callbacks["resourceNames"] = luaFunction {
		returnBuffer.setTo(tableOf(self.resources.keys.toTypedArray()))
	}

	callbacks["isResource"] = luaFunction { name: ByteString ->
		returnBuffer.setTo(name.decode() in self.resources.keys)
	}

	callbacks["resource"] = luaFunction { name: ByteString ->
		val resource = self.resources[name.decode()] ?: return@luaFunction returnBuffer.setTo(0.0)
		returnBuffer.setTo(resource.value)
	}

	callbacks["resourcePositive"] = luaFunction { name: ByteString ->
		val resource = self.resources[name.decode()] ?: return@luaFunction returnBuffer.setTo(false)
		returnBuffer.setTo(resource.value > 0.0)
	}

	callbacks["setResource"] = luaFunction { name: ByteString, value: Number ->
		val resource = self.resources[name.decode()] ?: throw LuaRuntimeException("No such resource $name")
		resource.value = value.toDouble()
	}

	callbacks["modifyResource"] = luaFunction { name: ByteString, value: Number ->
		val resource = self.resources[name.decode()] ?: throw LuaRuntimeException("No such resource $name")
		resource.value += value.toDouble()
	}

	callbacks["giveResource"] = luaFunction { name: ByteString, value: Number ->
		// while other functions throw an exception if resource does not exist
		// this one returns 0
		// Consistency is my first, second, and last name
		val resource = self.resources[name.decode()] ?: return@luaFunction returnBuffer.setTo(0.0)
		returnBuffer.setTo(resource.give(value.toDouble()))
	}

	callbacks["consumeResource"] = luaFunction { name: ByteString, value: Number ->
		// while other functions throw an exception if resource does not exist
		// this one returns 0
		// Consistency is my first, second, and last name
		val resource = self.resources[name.decode()] ?: return@luaFunction returnBuffer.setTo(false)
		returnBuffer.setTo(resource.consume(value.toDouble()))
	}

	callbacks["overConsumeResource"] = luaFunction { name: ByteString, value: Number ->
		// while other functions throw an exception if resource does not exist
		// this one returns 0
		// Consistency is my first, second, and last name
		val resource = self.resources[name.decode()] ?: return@luaFunction returnBuffer.setTo(false)
		returnBuffer.setTo(resource.consume(value.toDouble(), allowOverdraw = true))
	}

	callbacks["resourceLocked"] = luaFunction { name: ByteString ->
		val resource = self.resources[name.decode()] ?: throw LuaRuntimeException("No such resource $name")
		returnBuffer.setTo(resource.isLocked)
	}

	callbacks["setResourceLocked"] = luaFunction { name: ByteString, isLocked: Boolean ->
		val resource = self.resources[name.decode()] ?: throw LuaRuntimeException("No such resource $name")
		resource.isLocked = isLocked
	}

	callbacks["resetResource"] = luaFunction { name: ByteString, isLocked: Boolean ->
		val resource = self.resources[name.decode()] ?: throw LuaRuntimeException("No such resource $name")
		resource.reset()
	}

	callbacks["resetAllResources"] = luaFunction { name: ByteString, isLocked: Boolean ->
		self.resetResources()
	}

	callbacks["resourceMax"] = luaFunction { name: ByteString ->
		val resource = self.resources[name.decode()] ?: throw LuaRuntimeException("No such resource $name")
		returnBuffer.setTo(resource.maxValue)
	}

	callbacks["resourcePercentage"] = luaFunction { name: ByteString ->
		val resource = self.resources[name.decode()] ?: throw LuaRuntimeException("No such resource $name")
		returnBuffer.setTo(resource.percentage)
	}

	callbacks["setResourcePercentage"] = luaFunction { name: ByteString, value: Number ->
		val resource = self.resources[name.decode()] ?: throw LuaRuntimeException("No such resource $name")
		resource.setAsPercentage(value.toDouble())
	}

	callbacks["modifyResourcePercentage"] = luaFunction { name: ByteString, value: Number ->
		val resource = self.resources[name.decode()] ?: throw LuaRuntimeException("No such resource $name")
		resource.modifyPercentage(value.toDouble())
	}

	callbacks["getPersistentEffects"] = luaFunction { category: ByteString ->
		returnBuffer.setTo(tableOf(self.getPersistentEffects(category.decode()).map { it.map({ from(Starbound.gson.toJsonTree(it)) }, { it }) }))
	}

	callbacks["addPersistentEffect"] = luaFunction { category: ByteString, effect: Any ->
		self.addPersistentEffect(category.decode().sbIntern(), Starbound.gson.fromJsonFast(toJsonFromLua(effect), PersistentStatusEffectToken))
	}

	callbacks["addPersistentEffects"] = luaFunction { category: ByteString, effect: Any ->
		self.addPersistentEffects(category.decode().sbIntern(), Starbound.gson.fromJsonFast(toJsonFromLua(effect), CPersistentStatusEffectToken))
	}

	callbacks["setPersistentEffects"] = luaFunction { category: ByteString, effect: Any ->
		self.setPersistentEffects(category.decode().sbIntern(), Starbound.gson.fromJsonFast(toJsonFromLua(effect), CPersistentStatusEffectToken))
	}

	callbacks["clearPersistentEffects"] = luaFunction { category: ByteString ->
		self.removePersistentEffects(category.decode())
	}

	callbacks["clearAllPersistentEffects"] = luaFunction {
		self.removeAllPersistentEffects()
	}

	callbacks["addEphemeralEffect"] = luaFunction { name: ByteString, duration: Number?, source: Number? ->
		self.addEphemeralEffect(EphemeralStatusEffect(Registries.statusEffects.ref(name.decode().sbIntern()), duration?.toDouble()), source?.toInt())
	}

	callbacks["addEphemeralEffects"] = luaFunction { effects: Table, source: Number? ->
		for ((_, effect) in effects) {
			self.addEphemeralEffect(Starbound.gson.fromJsonFast(toJsonFromLua(effect), EphemeralStatusEffect::class.java), source?.toInt())
		}
	}

	callbacks["removeEphemeralEffect"] = luaFunction { name: ByteString ->
		self.removeEphemeralEffect(name.decode())
	}

	callbacks["clearEphemeralEffects"] = luaFunction {
		self.removeEphemeralEffects()
	}

	callbacks["damageTakenSince"] = luaFunction { since: Number? ->
		val (list, newSince) = self.recentDamageReceived(since?.toLong() ?: 0L)
		returnBuffer.setTo(tableOf(*list.map { from(Starbound.gson.toJsonTree(it)) }.toTypedArray()), newSince)
	}

	callbacks["inflictedHitsSince"] = luaFunction { since: Number? ->
		val (list, newSince) = self.recentHitsDealt(since?.toLong() ?: 0L)

		returnBuffer.setTo(tableOf(*list.map { p ->
			from(Starbound.gson.toJsonTree(p.second).also { it as JsonObject; it["targetEntityId"] = p.first })
		}.toTypedArray()), newSince)
	}

	callbacks["inflictedDamageSince"] = luaFunction { since: Number? ->
		val (list, newSince) = self.recentDamageDealt(since?.toLong() ?: 0L)
		returnBuffer.setTo(tableOf(*list.map { from(Starbound.gson.toJsonTree(it)) }.toTypedArray()), newSince)
	}

	callbacks["activeUniqueStatusEffectSummary"] = luaFunction {
		returnBuffer.setTo(tableOf(*self.activeUniqueStatusEffectSummary().map { tableOf(it.first.key, it.second) }.toTypedArray()))
	}

	callbacks["uniqueStatusEffectActive"] = luaFunction { name: ByteString ->
		returnBuffer.setTo(self.uniqueStatusEffectActive(name.decode()))
	}

	callbacks["primaryDirectives"] = luaFunction {
		returnBuffer.setTo(self.primaryDirectives.toByteString())
	}

	callbacks["setPrimaryDirectives"] = luaFunction { directives: ByteString? ->
		self.primaryDirectives = directives?.decode()?.sbIntern() ?: ""
	}

	callbacks["applySelfDamageRequest"] = luaFunction { damage: Table ->
		self.inflictSelfDamage(Starbound.gson.fromJsonFast(toJsonFromLua(damage), DamageData::class.java))
	}

	callbacks["appliesEnvironmentStatusEffects"] = luaFunction {
		returnBuffer.setTo(self.appliesEnvironmentStatusEffects)
	}

	callbacks["appliesWeatherStatusEffects"] = luaFunction {
		returnBuffer.setTo(self.appliesWeatherStatusEffects)
	}

	callbacks["minimumLiquidStatusEffectPercentage"] = luaFunction {
		returnBuffer.setTo(self.minimumLiquidStatusEffectPercentage)
	}

	callbacks["setAppliesEnvironmentStatusEffects"] = luaFunction { should: Boolean ->
		self.appliesEnvironmentStatusEffects = should
	}

	callbacks["setAppliesWeatherStatusEffects"] = luaFunction { should: Boolean ->
		self.appliesWeatherStatusEffects = should
	}

	callbacks["setMinimumLiquidStatusEffectPercentage"] = luaFunction { value: Number ->
		self.minimumLiquidStatusEffectPercentage = value.toDouble()
	}
}
