package ru.dbotthepony.kstarbound.world.entities

import com.google.common.collect.ImmutableList
import com.google.common.collect.ImmutableMap
import com.google.common.collect.ImmutableSet
import com.google.gson.JsonArray
import com.google.gson.JsonElement
import com.google.gson.JsonNull
import com.google.gson.JsonObject
import it.unimi.dsi.fastutil.objects.ObjectArraySet
import org.classdump.luna.ByteString
import org.classdump.luna.Table
import ru.dbotthepony.kommons.gson.get
import ru.dbotthepony.kommons.gson.set
import ru.dbotthepony.kstarbound.io.map
import ru.dbotthepony.kommons.util.Either
import ru.dbotthepony.kommons.util.getValue
import ru.dbotthepony.kommons.util.setValue
import ru.dbotthepony.kommons.util.value
import ru.dbotthepony.kstarbound.Registries
import ru.dbotthepony.kstarbound.Registry
import ru.dbotthepony.kstarbound.Starbound
import ru.dbotthepony.kstarbound.defs.ActorMovementParameters
import ru.dbotthepony.kstarbound.defs.DamageNotification
import ru.dbotthepony.kstarbound.defs.DamageSource
import ru.dbotthepony.kstarbound.defs.EntityDamageTeam
import ru.dbotthepony.kstarbound.defs.EntityType
import ru.dbotthepony.kstarbound.defs.HitType
import ru.dbotthepony.kstarbound.defs.InteractAction
import ru.dbotthepony.kstarbound.defs.InteractRequest
import ru.dbotthepony.kstarbound.defs.JumpProfile
import ru.dbotthepony.kstarbound.defs.PhysicsForceRegion
import ru.dbotthepony.kstarbound.defs.actor.StatModifier
import ru.dbotthepony.kstarbound.defs.actor.StatModifierType
import ru.dbotthepony.kstarbound.defs.actor.StatusControllerConfig
import ru.dbotthepony.kstarbound.defs.item.TreasurePoolDefinition
import ru.dbotthepony.kstarbound.defs.monster.MonsterVariant
import ru.dbotthepony.kstarbound.fromJsonFast
import ru.dbotthepony.kstarbound.io.DoubleValueCodec
import ru.dbotthepony.kstarbound.io.FloatValueCodec
import ru.dbotthepony.kstarbound.io.nullable
import ru.dbotthepony.kstarbound.json.builder.JsonFactory
import ru.dbotthepony.kstarbound.lua.LuaEnvironment
import ru.dbotthepony.kstarbound.lua.LuaMessageHandlerComponent
import ru.dbotthepony.kstarbound.lua.LuaUpdateComponent
import ru.dbotthepony.kstarbound.lua.bindings.MovementControllerBindings
import ru.dbotthepony.kstarbound.lua.bindings.provideAnimatorBindings
import ru.dbotthepony.kstarbound.lua.bindings.provideConfigBindings
import ru.dbotthepony.kstarbound.lua.bindings.provideEntityBindings
import ru.dbotthepony.kstarbound.lua.from
import ru.dbotthepony.kstarbound.lua.get
import ru.dbotthepony.kstarbound.lua.set
import ru.dbotthepony.kstarbound.lua.tableMapOf
import ru.dbotthepony.kstarbound.lua.toJsonFromLua
import ru.dbotthepony.kstarbound.lua.userdata.BehaviorState
import ru.dbotthepony.kstarbound.math.AABB
import ru.dbotthepony.kstarbound.math.vector.Vector2d
import ru.dbotthepony.kstarbound.network.packets.DamageRequestPacket
import ru.dbotthepony.kstarbound.network.syncher.InternedStringCodec
import ru.dbotthepony.kstarbound.network.syncher.JsonElementCodec
import ru.dbotthepony.kstarbound.network.syncher.NetworkedList
import ru.dbotthepony.kstarbound.network.syncher.NetworkedMap
import ru.dbotthepony.kstarbound.network.syncher.networkedBoolean
import ru.dbotthepony.kstarbound.network.syncher.networkedData
import ru.dbotthepony.kstarbound.network.syncher.networkedEnum
import ru.dbotthepony.kstarbound.network.syncher.networkedEventCounter
import ru.dbotthepony.kstarbound.network.syncher.networkedJsonElement
import ru.dbotthepony.kstarbound.network.syncher.networkedString
import ru.dbotthepony.kstarbound.util.random.MWCRandom
import ru.dbotthepony.kstarbound.util.random.random
import ru.dbotthepony.kstarbound.world.Direction
import ru.dbotthepony.kstarbound.world.World
import ru.dbotthepony.kstarbound.world.entities.api.InteractiveEntity
import ru.dbotthepony.kstarbound.world.entities.api.ScriptedEntity
import ru.dbotthepony.kstarbound.world.physics.Poly
import java.io.DataOutputStream

class MonsterEntity(val variant: MonsterVariant, level: Double? = null) : ActorEntity(), ScriptedEntity, InteractiveEntity {
	override val type: EntityType
		get() = EntityType.MONSTER

	init {
		team.accept(EntityDamageTeam(variant.commonParameters.damageTeamType, variant.commonParameters.damageTeam))
	}

	val lua = LuaEnvironment()
	val luaUpdate = LuaUpdateComponent(lua)
	val luaMovement = MovementControllerBindings(movement)
	val luaMessages = LuaMessageHandlerComponent(lua) { toString() }
	val animator = Animator(variant.animationConfig)

	init {
		lua.globals["storage"] = lua.newTable()
	}

	override fun move(delta: Double) {
		luaMovement.apply()
		super.move(delta)
	}

	init {
		for ((k, v) in variant.animatorPartTags) {
			animator.setPartTag(k, "partImage", v)
		}

		animator.zoom = variant.commonParameters.scale

		val colorSwap = variant.commonParameters.colorSwap
			?: Registries.monsterPalettes[variant.parameters.get("colors", "default")]?.value?.swaps?.random(MWCRandom(variant.seed.toULong(), 256, 32))

		if (colorSwap != null) {
			animator.processingDirectives = colorSwap.toImageOperator()
		}

		movement.applyParameters(variant.commonParameters.movementSettings)

		movement.applyParameters(ActorMovementParameters(
			standingPoly = variant.commonParameters.movementSettings.standingPoly?.flatMap({ it * variant.commonParameters.scale }, { it.stream().map { it * variant.commonParameters.scale }.collect(ImmutableList.toImmutableList()) }),
			crouchingPoly = variant.commonParameters.movementSettings.crouchingPoly?.flatMap({ it * variant.commonParameters.scale }, { it.stream().map { it * variant.commonParameters.scale }.collect(ImmutableList.toImmutableList()) }),
			walkSpeed = (movement.actorMovementParameters.walkSpeed ?: 0.0) * variant.commonParameters.walkMultiplier,
			runSpeed = (movement.actorMovementParameters.runSpeed ?: 0.0) * variant.commonParameters.runMultiplier,

			airJumpProfile = JumpProfile(
				jumpSpeed = (movement.actorMovementParameters.airJumpProfile.jumpSpeed ?: 0.0) * variant.commonParameters.jumpMultiplier,
			),

			liquidJumpProfile = JumpProfile(
				jumpSpeed = (movement.actorMovementParameters.liquidJumpProfile.jumpSpeed ?: 0.0) * variant.commonParameters.jumpMultiplier,
			),

			mass = (movement.actorMovementParameters.mass ?: 0.0) * variant.commonParameters.weightMultiplier,

			physicsEffectCategories = if (movement.actorMovementParameters.physicsEffectCategories == null) ImmutableSet.of("monster") else null,
		))

		networkGroup.upstream.add(uniqueID)
		networkGroup.upstream.add(team)
	}

	@JsonFactory
	data class SerializedData(
		val movementState: ActorMovementController.SerializedData,
		val statusController: StatusController.SerializedData,
		val monsterLevel: Double? = null,
		val damageOnTouch: Boolean,
		val aggressive: Boolean,
		val deathParticleBurst: String,
		val deathSound: String,
		val activeSkillName: String,
		val dropPool: Either<ImmutableMap<String, Registry.Ref<TreasurePoolDefinition>>, Registry.Ref<TreasurePoolDefinition>>? = null,
		val uniqueId: String? = null,
		val team: EntityDamageTeam,
		val scriptStorage: JsonElement = JsonNull.INSTANCE,
		val effectEmitter: JsonElement = JsonNull.INSTANCE,
	)

	override fun deserialize(data: JsonObject) {
		super.deserialize(data)

		val deserialized = Starbound.gson.fromJsonFast(data, SerializedData::class.java)
		movement.deserialize(deserialized.movementState)
		statusController.deserialize(deserialized.statusController)
		this.monsterLevel = deserialized.monsterLevel
		this.damageOnTouch = deserialized.damageOnTouch
		this.isAggressive = deserialized.aggressive
		this.deathParticlesBurst = deserialized.deathParticleBurst
		this.deathSound = deserialized.deathSound
		this.activeSkillName = deserialized.activeSkillName
		this.dropPool = deserialized.dropPool
		this.uniqueID.value = deserialized.uniqueId
		this.team.value = deserialized.team

		this.lua.globals["storage"] = lua.from(deserialized.scriptStorage)
	}

	override fun serialize(data: JsonObject) {
		super.serialize(data)

		val moreData = SerializedData(
			movementState = movement.serialize(),
			statusController = statusController.serialize(),
			monsterLevel = monsterLevel,
			damageOnTouch = damageOnTouch,
			aggressive = isAggressive,
			deathParticleBurst = deathParticlesBurst,
			deathSound,
			activeSkillName,
			dropPool,
			uniqueID.value,
			team.value,
			effectEmitter = effectEmitter.serialize(),
			scriptStorage = toJsonFromLua(lua.globals["storage"]),
		)

		for ((k, v) in Starbound.gson.toJsonTree(moreData).asJsonObject.entrySet()) {
			data[k] = v
		}

		data["monsterVariant"] = Starbound.gson.toJsonTree(variant)
	}

	var monsterLevel by networkedData(level, DoubleValueCodec.nullable(), FloatValueCodec.map({ toDouble() }, { toFloat() }).nullable()).also { networkGroup.upstream.add(it) }
		private set
	var damageOnTouch by networkedBoolean().also { networkGroup.upstream.add(it) }

	val customDamageSources = NetworkedList(DamageSource.CODEC, DamageSource.LEGACY_CODEC).also { networkGroup.upstream.add(it) }

	var dropPool = variant.chosenDropPool
	var isAggressive by networkedBoolean().also { networkGroup.upstream.add(it) }
	var knockedOut by networkedBoolean().also { networkGroup.upstream.add(it) }
	var knockoutTimer = 0.0
		private set

	var deathParticlesBurst by networkedString().also { networkGroup.upstream.add(it) }
	var deathSound by networkedString().also { networkGroup.upstream.add(it) }
	var activeSkillName by networkedString().also { networkGroup.upstream.add(it) }
	var networkName by networkedData(null, InternedStringCodec.nullable()).also { networkGroup.upstream.add(it) }

	var displayNameTag by networkedBoolean().also { networkGroup.upstream.add(it) }

	init {
		// why the fuck is this networked
		// anyway.
		/* val dropPool =*/ networkedJsonElement().also { networkGroup.upstream.add(it) }
	}

	val forceRegions = NetworkedList(PhysicsForceRegion.CODEC, PhysicsForceRegion.LEGACY_CODEC).also { networkGroup.upstream.add(it) }
	override val statusController: StatusController = StatusController(this, variant.commonParameters.statusSettings ?: StatusControllerConfig.EMPTY)

	init {
		networkGroup.upstream.add(animator.networkGroup)
		networkGroup.upstream.add(movement.networkGroup)
		networkGroup.upstream.add(statusController)
	}

	val effectEmitter = EffectEmitter(this).also { networkGroup.upstream.add(it.networkGroup) }

	val newChatMessageEvent = networkedEventCounter().also { networkGroup.upstream.add(it) }
	var chatMessage by networkedString().also { networkGroup.upstream.add(it) }
	var chatPortrait by networkedString().also { networkGroup.upstream.add(it) }

	// despite being networked as data on original engine, it is 1 byte wide
	override var damageBarType: DamageBarType by networkedEnum(DamageBarType.DEFAULT).also { networkGroup.upstream.add(it) }
	override var isInteractive by networkedBoolean().also { networkGroup.upstream.add(it) }

	// don't interpolate scripted animation parameters or animationdamageparts
	val animationDamageParts = NetworkedList(InternedStringCodec).also { networkGroup.upstream.add(it, propagateInterpolation = false) }
	val scriptedAnimationParameters = NetworkedMap(InternedStringCodec, JsonElementCodec).also { networkGroup.upstream.add(it, propagateInterpolation = false) }

	fun addChatMessage(message: String, portrait: String? = null) {
		chatMessage = message
		chatPortrait = portrait ?: ""
		newChatMessageEvent.trigger()
	}

	override var isPersistent: Boolean = variant.commonParameters.persistent

	override fun writeNetwork(stream: DataOutputStream, isLegacy: Boolean) {
		variant.write(stream, isLegacy)
	}

	override fun onJoinWorld(world: World<*, *>) {
		super.onJoinWorld(world)

		monsterLevel = monsterLevel ?: world.template.threatLevel

		if (!isRemote) {
			val healthMultiplier = (variant.commonParameters.healthLevelFunction.value?.evaluate(monsterLevel!!) ?: 1.0) * variant.commonParameters.healthMultiplier
			statusController.setPersistentEffects("innate", listOf(Either.left(StatModifier("maxHealth", healthMultiplier, StatModifierType.BASE_MULTIPLICATION))))

			provideEntityBindings(this, lua)
			provideAnimatorBindings(animator, lua)

			provideConfigBindings(lua) { key, default ->
				key.find(variant.parameters) ?: default
			}

			lua.attach(variant.commonParameters.scripts)
			luaUpdate.stepCount = variant.commonParameters.initialScriptDelta

			BehaviorState.provideBindings(lua)

			luaMovement.init(lua)
			lua.init()
		}
	}

	override val metaBoundingBox: AABB
		get() = variant.commonParameters.metaBoundBox + position

	override val mouthPosition: Vector2d
		get() = movement.getAbsolutePosition(variant.commonParameters.mouthOffset)

	override val feetPosition: Vector2d
		get() = movement.getAbsolutePosition(variant.commonParameters.feetOffset)

	override val name: String
		get() = networkName ?: variant.shortDescription ?: ""

	override val description: String
		get() = variant.description ?: "Some indescribable horror"

	override fun queryHit(source: DamageSource, attacker: AbstractEntity?, inflictor: AbstractEntity?): HitType? {
		if (knockedOut || statusController.statPositive("invulnerable"))
			return null

		if (source.intersect(world.geometry, damageHitbox))
			return HitType.HIT

		return null
	}

	override val damageHitbox: List<Poly>
		get() = listOf(variant.commonParameters.selfDamagePoly.rotate(movement.rotation) + position)

	private val deathDamageKinds = ObjectArraySet<String>()

	override fun experienceDamage(damage: DamageRequestPacket): List<DamageNotification> {
		val notifications = statusController.experienceDamage(damage.request)
		val totalDamage = notifications.sumOf { it.healthLost }

		if (totalDamage > 0.0) {
			lua.invokeGlobal("damage", lua.tableMapOf(
				"sourceId" to damage.request.sourceEntityId,
				"damage" to totalDamage,
				"sourceDamage" to damage.request.damage,
				"sourceKind" to damage.request.damageSourceKind
			))
		}

		if (health <= 0.0) {
			deathDamageKinds.add(damage.request.damageSourceKind)
		}

		return notifications
	}

	private val shouldDie: Boolean get() {
		val result = lua.invokeGlobal("shouldDie")
		return result.isNotEmpty() && result[0] is Boolean && result[0] as Boolean || health <= 0.0 //|| lua.errorState
	}

	override fun tick(delta: Double) {
		super.tick(delta)

		if (isLocal) {
			animator.isFlipped = (movement.facingDirection == Direction.LEFT) != variant.reversed

			if (knockedOut) {
				knockoutTimer -= delta

				if (knockoutTimer <= 0.0) {
					val dropPool = if (dropPool == null) null else if (dropPool!!.isRight) {
						dropPool!!.right()
					} else {
						// Check to see whether any of the damage types that were used to cause
						// death are in the damage pool map, if so spawn treasure from that,
						// otherwise set the treasure pool to the "default" entry.

						val map = dropPool!!.left()
						deathDamageKinds.firstNotNullOfOrNull { map[it] } ?: map["default"]
					}

					if (dropPool?.value != null) {
						for (stack in dropPool.value!!.evaluate(world.random, monsterLevel ?: world.template.threatLevel)) {
							val entity = ItemDropEntity(stack)
							entity.movement.position = movement.computeGlobalHitboxes().stream().map { it.centre }.reduce { t, u -> (t + u) / 2.0 }.orElse(position)
							entity.movement.velocity = movement.velocity
							entity.joinWorld(world)
						}
					}

					remove(RemovalReason.DYING)
					return
				}
			} else {
				luaUpdate.update(delta) {
					luaMovement.clearControlsIfNeeded()
					forceRegions.clear()
				}

				if (shouldDie) {
					knockedOut = true
					knockoutTimer = variant.commonParameters.knockoutTime
					damageOnTouch = false

					if (!variant.commonParameters.knockoutEffect.isNullOrBlank()) {
						animator.setEffectActive(variant.commonParameters.knockoutEffect, true)
					}

					for ((k, v) in variant.commonParameters.knockoutAnimationStates) {
						animator.setActiveState(k, v)
					}
				}
			}
		}

		effectEmitter.setSourcePosition("normal", position)
		effectEmitter.setSourcePosition("mouth", mouthPosition)
		effectEmitter.setSourcePosition("feet", mouthPosition)
		effectEmitter.direction = movement.facingDirection
		effectEmitter.tick(isRemote, visibleToRemotes)

		if (world.isServer) {
			animator.tick(delta, world.random)
		}
	}

	override fun handleMessage(connection: Int, message: String, arguments: JsonArray): JsonElement? {
		return luaMessages.handle(message, connection == connectionID, arguments) ?: statusController.handleMessage(message, connection == connectionID, arguments)
	}

	override fun callScript(fnName: String, vararg arguments: Any?): Array<Any?> {
		return lua.invokeGlobal(fnName, *arguments)
	}

	override fun evalScript(code: String): Array<Any?> {
		return lua.eval(code)
	}

	override fun interact(request: InteractRequest): InteractAction {
		val result = lua.invokeGlobal("interact", lua.tableMapOf(
			"sourceId" to request.source,
			"sourcePosition" to lua.from(request.sourcePos)
		))

		if (result.isEmpty() || result[0] == null)
			return InteractAction.NONE

		val value = result[0]

		if (value is ByteString)
			return InteractAction(value.decode(), entityID)

		value as Table
		return InteractAction((value[1L] as ByteString).decode(), entityID, toJsonFromLua(value[2L]))
	}
}
